"""
life-simplifiyng stuff for lazy devs
--------
"""



__title__ = "team4958_customs"
__author__ = "arti404"
__license__ = "MIT"
__copyright__ = "Copyright 2023-present arti404"
__version__ = "0.0.0"



from team4958_customs import tools, utilities
from team4958_customs.utils import MISSING, ROOT



__all__ = [
    'tools',
    'utilities',
    'MISSING',
    'ROOT'
]


"""
if you are tired of recreating same project structures again and again
--------
"""



from team4958_customs.utilities.assets import builder_presets
from team4958_customs.utils import ROOT, MISSING, MAXINT, _yesornot

import pathlib
from pathlib import Path

import os


__all__ = [
    'Build'
]



NUMBERS = ['1', '2', '3', '4', '5', '6', '7', '8', '9', '0']
LOWERCASES = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z']
UPPERCASES = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z']
ALLOWED_SYM = ['_']
UNRECOMMENDED_SYM = [' ','(',')','-',"'",';']

RESTRICTED = ['`', '~', '!', '@', '#', '$','^','&','*','=','+','?','{','}',':','<',',','.','>','№','|','/','\\','"']
ALLOWED = NUMBERS+LOWERCASES+ALLOWED_SYM
UNRECOMMENDED = UPPERCASES+UNRECOMMENDED_SYM



class _Check:
    def sym_presence(string:str, sym_list:list):
        """
        checks if symbols from listing are present in string
        """
        res = []
        string = list(set(list(string)))
        for sym in string:
            if sym in sym_list:
                res.append(sym)
        if res!=[]:
            return res
        else:
            return



def _projectname():
    from sys import setrecursionlimit as recurlim
    recurlim(MAXINT)
    name = input("give your project a proper name: ")
    restricted = _Check.sym_presence(name, RESTRICTED)
    unrecommended = _Check.sym_presence(name, UNRECOMMENDED)
    if name=='':
        print('name cannot be empty!')
        return _projectname()
    elif restricted:
        print(f"there are some restricted symbols in project name given: {restricted}")
        return _projectname()
    elif unrecommended:
        confirm = _yesornot(f'there are some unrecommended symbols in project name!\nsymbols checker didnt like: {unrecommended}\ndo you really wish to keep it like that?')
        if confirm:
            return name
        else:
            return _projectname()
    else:
        return name



def _fromdict(obj:dict, path=MISSING):
        """
        `NEVER FUCKING CALL THIS INSIDE YOUR CODE`
        ------------
        \n
        this will spawn your whole structure directly in current work directory!\n
        `use 'fromdict' method instead`\n
        --------\n
        builds a project structure from dict
        ----
        \n
        -------\n
        example_dict = {
            'folder_name':{
                #feel free to make as much substructures as you need
            },
            'filename.extension':['line1', 'line2', 'line3', etc.],
            'filename2.extension':'any text to put inside the file just as string'
        }
        """
        if path is MISSING:
            path = ROOT
        for (key, val) in obj.items():
            if type(val) is dict:
                try:
                    os.mkdir(pathlib.Path(path, key))
                except FileExistsError:
                    pass
                newpath = pathlib.Path(path, key)
                _fromdict(val, newpath)
            elif type(val) is list or type(val) is str:
                with open(pathlib.Path(path, key), 'w') as f:
                    if type(val) is list:
                        if val!=[]:
                            for line in val:
                                f.write(f'{line}\n')
                    else:
                        f.write(val)
            else:
                print(f'"{key}" type not specified as file or folder, skipping...')



class Build:
    """builds a project template from preset inside a subfolder named as your project"""

    def blank(name: str=MISSING, create_subfolder=True, readme=True, requirements=True, license_file=False):
        """
        blank preset which creates `src` folder and `main.py` file without anything inside
        ---------
        \n
        also creates empty `README.md`, `requirements.txt` and `LICENSE` files\n
        ---------\n
        name: name of your project folder (if not specified inside the code will be asked through console)\n
        required when create_subfolder is 'True' (which is deafult)\n
        --------------\n
        create_subfolder: whether you wish to create a subfolder for your project (builds template in a current directory if 'False')\n
        -------\n
        readme: whether you wish to create a `README.md` file in your project's root directory\n
        -------\n
        requirements: whether you wish to create a `requirements.txt` file in your project's root directory\n
        ---------\n
        license_file: whether you wish to create a `LICENSE` file in your project's root directory (will be empty)
        """
        if create_subfolder == False:
            proj_path = ROOT
        else:
            if name is MISSING:
                name = _projectname()
                try:
                    os.mkdir(Path(ROOT, name))
                except FileExistsError:
                    pass
                proj_path = Path(ROOT, name)
            else:
                try:
                    os.mkdir(Path(ROOT, name))
                except FileExistsError:
                    pass
                proj_path = Path(ROOT, name)
        with open(Path(proj_path, 'main.py'), 'w') as f:
            pass
        if readme:
            with open(Path(proj_path, 'README.md'), 'w') as f:
                pass
        if requirements:
            with open(Path(proj_path, 'requirements.txt'), 'w') as f:
                f.write("team4958_customs")
        if license_file:
            with open(Path(proj_path, 'LICENSE'), 'w') as f:
                pass
    
    def fromdict(preset:dict, name: str=MISSING, create_subfolder=True):
        """
        builds a project structure from dict
        ----
        \n
        ---------\n
        preset: your own structure preset presented as dictionary
        \n
        example_dict = {\n
            'folder_name':{\n
                #feel free to make as much substructures as you need\n
            },\n
            'filename.extension':['line1', 'line2', 'line3', etc.],\n
            'filename2.extension':'any text to put inside the file just as string'\n
        }\n
        -------\n
        name: your project's name (required and has its point only if 'create_subfolder' is True)\n
        create_subfolder: whether you wish to create everything inside a subfolder\n
        """
        if create_subfolder == False:
            proj_path = ROOT
        else:
            if name is MISSING:
                name = _projectname()
                try:
                    os.mkdir(Path(ROOT, name))
                except FileExistsError:
                    pass
                proj_path = Path(ROOT, name)
            else:
                try:
                    os.mkdir(Path(ROOT, name))
                except FileExistsError:
                    pass
                proj_path = Path(ROOT, name)
        _fromdict(preset, path=proj_path)
    
    def deafult(name: str=MISSING, create_subfolder=True):
        """
        builds a project structure from built-in deafult template
        ----
        \n
        ---------\n
        name: your project's name (required and has its point only if 'create_subfolder' is True)\n
        create_subfolder: whether you wish to create everything inside a subfolder\n
        """
        if create_subfolder == False:
            proj_path = ROOT
        else:
            if name is MISSING:
                name = _projectname()
                try:
                    os.mkdir(Path(ROOT, name))
                except FileExistsError:
                    pass
                proj_path = Path(ROOT, name)
            else:
                try:
                    os.mkdir(Path(ROOT, name))
                except FileExistsError:
                    pass
                proj_path = Path(ROOT, name)
        _fromdict(builder_presets.DEAFULT, path=proj_path)
        
    def disnake_bot(name: str=MISSING, create_subfolder=True):
        """
        builds a disnake bot project structure from built-in template
        ----
        \n
        ---------\n
        name: your project's name (required and has its point only if 'create_subfolder' is True)\n
        create_subfolder: whether you wish to create everything inside a subfolder\n
        """
        if create_subfolder == False:
            proj_path = ROOT
        else:
            if name is MISSING:
                name = _projectname()
                try:
                    os.mkdir(Path(ROOT, name))
                except FileExistsError:
                    pass
                proj_path = Path(ROOT, name)
            else:
                try:
                    os.mkdir(Path(ROOT, name))
                except FileExistsError:
                    pass
                proj_path = Path(ROOT, name)
        _fromdict(builder_presets.DISNAKE, path=proj_path)



class Package:
    def __init__(self):
        pass

    def _pack():
        pass
    
    def _unpack():
        pass


DEAFULT = {
    'assets':{
        '__init__.py':''
    },
    'src':{
        '__init__.py':''
    },
    'README.md':[
        r"# -DEAFULT-PROJECT-",
        r"use it to make something incredible!"
    ],
    'main.py':r'import team4958_customs',
    'requirements.txt':''
}



DISNAKE = {
    'assets':{
        '__init__.py':''
    },
    'src':{
        'extensions':{
            'commands':{
                'blank':{
                    '__command__slash_cmd.py':[
                        r'"""common slash command template"""',
                        r'',
                        r'',
                        r'',
                        r'import disnake',
                        r'from disnake.ext import commands',
                        r'',
                        r'',
                        r'',
                        r'class slash_cmd(commands.Cog):',
                        r'',
                        r'    def __init__(self, bot):',
                        r'        self.bot = bot',
                        r'',
                        r'    @commands.slash_command(description="cmd description sample")',
                        r'    async def slash_cmd(self, inter):',
                        r'        await inter.send(f"slash-command sample", ephemeral=True)',
                        r'',
                        r'',
                        r'',
                        r'def setup(bot):',
                        r'    bot.add_cog(slash_cmd(bot))'
                    ],
                    '__init__.py':r'"""command templates"""'
                },
                '__init__.py':r'"""bot commands"""'
            },
            'events':{
                '__event__start.py':[
                    r'import logging',
                    r'',
                    r'from disnake.ext import commands',
                    r'',
                    r'',
                    r'',
                    r'log = logging.getLogger(__name__)',
                    r'',
                    r'',
                    r'',
                    r'class start(commands.Cog):',
                    r'',
                    r'    def __init__(self, bot):',
                    r'        self.bot = bot',
                    r'',
                    r'    @commands.Cog.listener()',
                    r'    async def on_ready(self):',
                    r"        log.critical(f'started {self.bot.user}')",
                    r"        print(f'started {self.bot.user}')",
                    r'',
                    r'',
                    r'',
                    r'def setup(bot):',
                    r'    bot.add_cog(start(bot))'
                ],
                '__init__.py':r'"""bot events to listen to"""'
            },
            '__init__.py':r'"""loadable bot extensions"""'
        },
        'utils':{
            'logging':{
                '__init__.py':'',
                'logger_setup.py':[
                    r'"""logging setup presets"""',
                    r'',
                    r'',
                    r'',
                    r'import logging',
                    r'from logging.handlers import TimedRotatingFileHandler',
                    r'',
                    r'import pathlib',
                    r'from pathlib import Path',
                    r'',
                    r'import os',
                    r'',
                    r'',
                    r'',
                    r'def normal(name):',
                    r'    """ALWAYS put __name__ as arg!!"""',
                    r'    if not os.path.isdir("logs"):',
                    r'        os.mkdir("logs")',
                    r"    file = Path(pathlib.Path.cwd(), 'logs', 'exc.log')",
                    r'''    logging.basicConfig(level=logging.WARNING, filename=file, filemode="a", format='%(asctime)s %(name)s %(levelname)s:%(message)s')''',
                    r'    logging.debug("A DEBUG Message")',
                    r'    logging.info("An INFO")',
                    r'    logging.warning("A WARNING")',
                    r'    logging.error("An ERROR")',
                    r'    logging.critical("A message of CRITICAL severity")',
                    r"    TimedRotatingFileHandler(filename=file, when='d', interval=1, backupCount=7)",
                    r'    return logging.getLogger(name)',
                    r'',
                    r'def debug(name):',
                    r'    """ALWAYS put __name__ as arg!!"""',
                    r'    if not os.path.isdir("logs"):',
                    r'        os.mkdir("logs")',
                    r"    file = Path(pathlib.Path.cwd(), 'logs', 'exc.log')",
                    r'''    logging.basicConfig(level=logging.DEBUG, filename=file, filemode="a", format='%(asctime)s %(name)s %(levelname)s:%(message)s')''',
                    r'    logging.debug("A DEBUG Message")',
                    r'    logging.info("An INFO")',
                    r'    logging.warning("A WARNING")',
                    r'    logging.error("An ERROR")',
                    r'    logging.critical("A message of CRITICAL severity")',
                    r"    TimedRotatingFileHandler(filename=file, when='d', interval=1, backupCount=7)",
                    r'    return logging.getLogger(name)',
                    r'',
                    r'def test(name):',
                    r'    """ALWAYS put __name__ as arg!!"""',
                    r'    if not os.path.isdir("logs"):',
                    r'        os.mkdir("logs")',
                    r"    file = Path(pathlib.Path.cwd(), 'logs', 'exc.log')",
                    r'''    logging.basicConfig(level=logging.INFO, filename=file, filemode="a", format='%(asctime)s %(name)s %(levelname)s:%(message)s')''',
                    r'    logging.debug("A DEBUG Message")',
                    r'    logging.info("An INFO")',
                    r'    logging.warning("A WARNING")',
                    r'    logging.error("An ERROR")',
                    r'    logging.critical("A message of CRITICAL severity")',
                    r"    TimedRotatingFileHandler(filename=file, when='d', interval=1, backupCount=7)",
                    r'    return logging.getLogger(name)'
                ]
            },
            '__init__.py':r'"""custom cog utils"""',
            'funcs.py':[
                r'"""custom python objects"""',
                r'',
                r'',
                r'',
                r'import logging',
                r'',
                r'',
                r'',
                r'log = logging.getLogger(__name__)',
                r'',
                r'',
                r'',
                r'#put some custom classes and functions here'
            ],
            'parsher.py':[
                r'"""deafult messages and embeds"""',
                r'',
                r'',
                r'',
                r'import disnake',
                r'',
                r'',
                r'',
                r'def create_embeds(emb_dicts):',
                r'    """creates a list of disnake embeds from list of dicts"""',
                r'    res = []',
                r'    for element in emb_dicts:',
                r'        embed = disnake.Embed.from_dict(element)',
                r'        res.append(embed)',
                r'    if len(res)>0:',
                r'        return res',
                r'    else:',
                r'        embed = disnake.Embed.from_dict({"title": "Embed Title","title": "Embed Title","color": 0xFEE75C})',
                r'        return [embed]',
                r'',
                r'',
                r'',
                r'"""embed dictionaries"""',
                r'"""',
                r'to set use:',
                r'    embed=disnake.Embed.from_dict(template)',
                r'"""',
                r'"""',
                r'embed_dict = {',
                r'    "title": "Embed Title",',
                r'   "description": "Embed Description",',
                r'    "color": 0xFEE75C,',
                r'    "timestamp": datetime.datetime.now().isoformat(),',
                r'    "author": {',
                r'        "name": "Embed Author",',
                r'        "url": "https://disnake.dev/",',
                r'        "icon_url": "https://disnake.dev/assets/disnake-logo.png",',
                r'    },',
                r'    "thumbnail": {"url": "https://disnake.dev/assets/disnake-logo.png"},',
                r'    "fields": [',
                r'        {"name": "Regular Title", "value": "Regular Value", "inline": "false"},',
                r'        {"name": "Inline Title", "value": "Inline Value", "inline": "true"},',
                r'        {"name": "Inline Title", "value": "Inline Value", "inline": "true"},',
                r'        {"name": "Inline Title", "value": "Inline Value", "inline": "true"},',
                r'    ],',
                r'    "image": {"url": "https://disnake.dev/assets/disnake-banner-thin.png"},',
                r'    "footer": {"text": "Embed Footer", "icon_url": "https://disnake.dev/assets/disnake-logo.png"},',
                r'}',
                r'"""'
            ]
        },
       '__init__.py':r'"""bot source code root folder"""'
    },
    'config.py':[
        r'TOKEN = "" #test bot token',
        r'',
        r'',
        r'',
        r'#DISCORD CONFIG',
        r"emb_image_none = 'https://cdn.discordapp.com/attachments/990257257986027550/1014942594448502894/5PAAPj7.png' #just an empty image",
        r'ds_col = 0x2f3136 #discord background color',
        r'guild_id = ... #discord guild id',
        r'server_pic = ... #server picture',
        r'',
        r'',
        r'',
        r'#your database config',
        r'db_config = {',
        r"    'mysql': { #bot's main db config",
        r"        'host': 'localhost',",
        r"        'user': '',",
        r"        'passwd': '',",
        r"        'user2': '',",
        r"        'passwd2': '',",
        r"        'database': '',",
        r"        'tables':[",
        r"            'table (column TYPE, column TYPE)'",
        r"        ]",
        r'    },',
        r"    'root_config': { #mysql server root user config",
        r"        'host': 'localhost',",
        r"        'user': 'root',",
        r"        'password': ''",
        r"    }",
        r'}',
        r'',
        r'',
        r'',
        r'#some common things',
        r"version = '0.0.0 (blank)'"
    ],
    'main.py':[
        r"import os",
        r'import sys',
        r'',
        r"if str(os.name)=='posix':",
        r'    os.system("python3 -m pip install -r requirements.txt")',
        r"elif str(os.name)=='nt':",
        r'    os.system("py -m pip install -r requirements.txt")',
        r'else:',
        r'    print("unknown os type, cannot launch")',
        r'    sys.exit()',
        r'',
        r'',
        r'',
        r'import traceback',
        r'from team4958_customs.utils import *',
        r'',
        r'',
        r'',
        r'def database():',
        r'    from team4958_customs.utils import _yesornot',
        r'    create = _yesornot("do you wish to create a database for your bot?")',
        r'    if create == True:',
        r'        from team4958_customs.tools.sql import Administration',
        r'        from config import db_config',
        r'        from mysql.connector import Error as sqlErr',
        r"        sql = Administration(db_config['mysql'], mysql_config=db_config['root_conf'])",
        r'        try:',
        r'            sql.create_db(add_tables=False)',
        r'        except sqlErr:',
        r'            print(f"failed to database!\n{traceback.print_exc()}")',
        r'        try:',
        r'            sql.create_tables()',
        r'        except sqlErr:',
        r'            print(f"failed to create tables in your database!\n{traceback.print_exc()}")',
        r'        try:',
        r'            sql.create_users()',
        r'        except sqlErr:',
        r'            print(f"failed to create users in your database!\n{traceback.print_exc()}")',
        r'database()',
        r'',
        r'',
        r'',
        r'from pathlib import Path',
        r'from src.utils.logging import logger_setup as logger',
        r'',
        r'',
        r'',
        r'def launch_mode():',
        r"    print('enter DEBUG mode? (y/n)', end='')",
        r"    mode = input()",
        r"    if mode == 'y' or mode == 'Y':",
        r'        print("starting debug mode...")',
        r"        output = Path(ROOT, 'logs', 'console.txt')",
        r"        sys.stdout = open(output, 'w')",
        r"        sys.stderr = open(output, 'w')",
        r"        return logger.debug(__name__)",
        r"    elif mode == 'n' or mode == 'N':",
        r'        print("starting normal...")',
        r"        output = Path(ROOT, 'logs', 'console.txt')",
        r"        sys.stdout = open(output, 'w')",
        r"        sys.stderr = open(output, 'w')",
        r'        return logger.normal(__name__)',
        r'    else:',
        r'        return logger.test(__name__)',
        r'log = launch_mode()',
        r'',
        r'',
        r'',
        r'import disnake',
        r'from disnake.ext import commands',
        r'from os import walk',
        r'from config import Token',
        r'',
        r'',
        r'',
        r'bot = commands.InteractionBot(intents= disnake.Intents.all())',
        r'',
        r'',
        r'',
        r"src_root = Path(ROOT, 'src', 'extensions')",
        r'loadlist = ["__command__", "__event__"]',
        r'def loadable(file):',
        r'    if file.endswith(".py"):',
        r'        for ext in loadlist:',
        r'            if file.startswith(ext):',
        r'                return True',
        r'    else:',
        r'        return',
        r'for (dirpath, dirnames, filenames) in walk(src_root):',
        r'    for filename in os.listdir(dirpath):',
        r'        if loadable(filename):',
        r'            path = Path(dirpath, filename)',
        r'            strpath = str(path)',
        r"            strpath = strpath.replace(str(Path(ROOT)), '')",
        r'            for char in [r"\\", r"\ "[0], "/", "//"]:',
        r'                if char in strpath:',
        r"                    ext = strpath.replace(char, '.')",
        r"                    ext = ext.replace(ext[0], '', 1)",
        r'            bot.load_extension(ext[:-3])',
        r'            print(f"extension {filename[:-3]} loaded!")',
        r'            log.warning(f"extension {filename[:-3]} loaded!")',
        r'',
        r'',
        r'',
        r'bot.run(Token)'
    ],
    'README.md':[
        r"# -Disnake-bot-BLANK-",
        r"template for disnake bot",
        r"",
        r"## get started",
        r"",
        r"log folder `'./logs'` will be created automatically on start if not exist!",
        r"",
        r"only files from `'./src/extensions'` folder and all its subfolders with `'__command__'` and `'__event__'` in their names *as prefix* are being loaded as disnake bot extensions!!"
    ],
    'requirements.txt':[
        r'disnake',
        r'team4958-customs'
    ]
}


from team4958_customs.utils import *

import asyncio
from asyncio import coroutines
import threading

import logging
import traceback

from inspect import isfunction
from typing import Iterable, Any



__all__ = [
    'Asyncio',
    'Threading'
]
log = logging.getLogger(__name__)
DEAFULT_LOOP = asyncio.get_event_loop()



async def _blank():
    pass


    
class Asyncio:
    """
    simplyfies work with `asyncio` which prvides you ability to call coroutines without 'await' expression\n
    -------\n
    loop: asyncio event loop to run your coroutine (highly recommended to specify for every instance uses 'run_threaded' method, but still optional); to set assign `asyncio.get_event_loop()` to a variable and give it\n
    pass_exc: whether you want to pass all catched exceptions by printing (and optionally logging) their tracebacks
    """
    def __init__(self, loop:asyncio.AbstractEventLoop=MISSING, pass_exc: bool=MISSING):
        self.loop = loop
        self.pass_exceptions = pass_exc
        
    def run(self, coro, pass_exceptions: bool=MISSING) -> Any:
        """
        runs your coroutine\n
        can return coroutine result\n
        don't use 'await' expression\n
        -------\n
        pass_exceptions: whether you want to pass all catched exceptions by printing (and optionally logging) their tracebacks
        """
        if not coroutines.iscoroutine(coro):
            raise TypeError("A coroutine object is required! To run a non-coro function use 'Threading' subclass instead.")
        if pass_exceptions is MISSING:
            if self.pass_exceptions is not MISSING:
                pass_exceptions = self.pass_exceptions
            else:
                pass_exceptions = False
        if pass_exceptions:
            try:
                return asyncio.run(coro)
            except Exception as err:
                log.error(f"passing exception during running coroutine\n{err}", exc_info=True)
                print("passing exception during running coroutine")
                traceback.print_exc()
        else:
            return asyncio.run(coro)
    
    def run_threaded(self, coro, loop:asyncio.AbstractEventLoop=MISSING):
        """
        runs your coroutine without blocking the rest code\n
        this method cannot return anything!\n
        also it always passes exceptions by just printing them\n
        can be used only in async functions\n
        -------\n
        loop: asyncio event loop to run your coroutine (unnecessary if already passed to class instance but may be overriden here)
        """
        if not coroutines.iscoroutine(coro):
            raise TypeError("A coroutine object is required! To run a non-coro function use 'Threading' class instead.")
        async def _exc_pass(obj):
            try:
                await obj
            except Exception as err:
                log.error(f"passing exception during running threaded coroutine\n{err}", exc_info=True)
                print("passing exception during running threaded coroutine")
                traceback.print_exc()
        if loop is MISSING:
            if self.loop is not MISSING:
                loop = self.loop
            else:
                loop = DEAFULT_LOOP
        asyncio.run_coroutine_threadsafe(_exc_pass(coro), loop)

    def kill(self, loop:asyncio.AbstractEventLoop=MISSING):
        """
        safely stops running loop\n
        affects the class instance loop if not specified\n
        if class instance doesn't nave a specified loop too, the deafult event loop will be affected
        """
        if loop is MISSING:
            if self.loop is not MISSING:
                loop = self.loop
            else:
                loop = DEAFULT_LOOP
        if loop.is_running:
            loop.stop()


    
class Threading:
    """
    simplyfies work with `threading` which prvides you ability to use non-coroutines asyncronously\n
    `functions threaded using this class are impossible to stop from code before they finish or stop by themselves!`\n
    also it's impossible to return anything from this\n
    ----------\n
    pass_exc: whether you want to pass all catched exceptions by printing (and optionally logging) their tracebacks
    """
    def __init__(self, pass_exc: bool=MISSING):
        self.pass_exceptions = pass_exc

    def run(self, func, args: Iterable=MISSING, pass_exceptions: bool=MISSING):
        """
        runs your non-coroutine function threaded\n
        `this method blocks your whole code from being safely stopped/killed until all threads are finished!`\n
        ---------\n
        func: your non-coroutine function (give only a function name here)\n
        args: arguments for function as `(arg1, arg2, ... ,)`\n
        pass_exceptions: whether you want to pass all catched exceptions by printing (and optionally logging) their tracebacks
        """
        if coroutines.iscoroutine(func):
            raise TypeError("A coroutine object is unacceptable! To run a coroutine function use 'Asyncio' subclass instead.")
        if not isfunction(func):
            raise TypeError("A function is required as 'func' argument!")
        if pass_exceptions is MISSING:
            if self.pass_exceptions is not MISSING:
                pass_exceptions = self.pass_exceptions
            else:
                pass_exceptions = False
        if pass_exceptions:
            try:
                if args is MISSING:
                    args = ()
                thr = threading.Thread(target=func, args=args)
                thr.start()
                thr.join()
            except Exception as err:
                log.error(f"passing exception during running threaded function\n{err}", exc_info=True)
                print("passing exception during running threaded function")
                traceback.print_exc()
        else:
            if args is MISSING:
                args = ()
            thr = threading.Thread(target=func, args=args)
            thr.start()
            thr.join()

    def run_daemon(self, func, args: Iterable=MISSING, pass_exceptions: bool=MISSING):
        """
        runs your non-coroutine function daemon-threaded\n
        `safe for stopping/killing your program anytime but never run critical tasks with this`\n
        ---------\n
        func: your non-coroutine function (give only a function name here)\n
        args: arguments for function\n
        pass_exceptions: whether you want to pass all catched exceptions with printing their tracebacks
        """
        if coroutines.iscoroutine(func):
            raise TypeError("A coroutine object is unacceptable! To run a coroutine function use 'Asyncio' subclass instead.")
        if not isfunction(func):
            raise TypeError("A function is required as 'func' argument!")
        if pass_exceptions is MISSING:
            if self.pass_exceptions is not MISSING:
                pass_exceptions = self.pass_exceptions
            else:
                pass_exceptions = False
        if pass_exceptions:
            try:
                if args is MISSING:
                    args = ()
                thr = threading.Thread(target=func, args=args, daemon=True)
                thr.start()
                thr.join()
            except Exception as err:
                log.error(f"passing exception during running threaded function\n{err}", exc_info=True)
                print("passing exception during running threaded function")
                traceback.print_exc()
        else:
            if args is MISSING:
                args = ()
            thr = threading.Thread(target=func, args=args, daemon=True)
            thr.start()
            thr.join()


"""
MySQL work simplification
---------
"""



import mysql.connector
from mysql.connector import Error as sqlErr

from team4958_customs.utils import *

import logging, traceback



__all__ = [
    'MySQLqueries',
    'Administration'
]
log = logging.getLogger(__name__)



class MySQLqueries():
    """
    basic one-srting MySQL queries
    -------
    \n
    works as `connect->execute query->commit if necessary->disconnect`\n
    helps to avoid freaking bunches of unused connections\n
    --------------------------\n
    dbconfig: your database config as `dict`{\n
        `'host'`: host address as str,\n
        `'user'`: db primary user username,\n
        `'passwd'`: db primary user password,\n
        `'user2'`: db secondary user username,\n
        `'passwd2'`: db secondary user password,\n
        `'database'`: your database name,\n
        `'tables'`: db tables structure presented as ['table (column TYPE, column TYPE, etc.)', etc.]\n
        }\n
    alt_user: specify if you need to set this parameter to whole class instance
    """

    def __init__(self, dbconfig:dict, alt_user=False):
        self.dbconfig = dbconfig
        self.alt_user=alt_user
    
    def commitable(self, sql_query:str, alt_user: bool=MISSING):
        """
        commit-requiring query\n
        has no output\n
        -----\n
        sql_query: MySQL query to
        ------\n
        alt_user: specify if you need to set for this call only
        """
        if alt_user is MISSING:
            alt_user=self.alt_user
        conn = BasicAction().connect(self.dbconfig, secondary=alt_user)
        if conn:
            cursor = conn.cursor()
            try:
                cursor.execute(sql_query)
                conn.commit()
            except sqlErr as err:
                log.error(f"failed commitable query: {sql_query}\n{err}", exc_info=True)
                print(f"failed commitable query: {sql_query}")
                traceback.print_exc()
            finally:
                cursor.close()
                conn.close()
    
    def fetch_one(self, sql_query:str, alt_user: bool=MISSING):
        """
        not commit-requiring query\n
        fetchone as output (returns tuple)\n
        -----\n
        alt_user: specify if you need to set for this call only
        """
        if alt_user is MISSING:
            alt_user=self.alt_user
        conn = BasicAction().connect(self.dbconfig, secondary=alt_user)
        if conn:
            cursor = conn.cursor()
            res = MISSING
            try:
                cursor.execute(sql_query)
                res = cursor.fetchone()
            except sqlErr as err:
                log.error(f"failed fetchable query: {sql_query}\n{err}", exc_info=True)
                print(f"failed fetchable query: {sql_query}")
                traceback.print_exc()
            finally:
                cursor.close()
                conn.close()
            return res
        else:
            return

    def fetch_all(self, sql_query:str, alt_user: bool=MISSING):
        """
        not commit-requiring query\n
        fetchall as output (returns list of tuples)\n
        -----\n
        alt_user: specify if you need to set for this call only
        """
        if alt_user is MISSING:
            alt_user=self.alt_user
        conn = BasicAction().connect(self.dbconfig, secondary=alt_user)
        if conn:
            cursor = conn.cursor()
            res = MISSING
            try:
                cursor.execute(sql_query)
                res = cursor.fetchall()
            except sqlErr as err:
                log.error(f"failed fetchable query: {sql_query}\n{err}", exc_info=True)
                print(f"failed fetchable query: {sql_query}")
                traceback.print_exc()
            finally:
                cursor.close()
                conn.close()
            return res
        else:
            return
    
    def fetch_many(self, sql_query:str, size:int, alt_user: bool=MISSING):
        """
        not commit-requiring query\n
        fetchmany as output (returns list of tuples)\n
        -----\n
        size: amount of elements to fetch from database\n
        alt_user: specify if you need to set for this call only
        """
        if alt_user is MISSING:
            alt_user=self.alt_user
        conn = BasicAction().connect(self.dbconfig, secondary=alt_user)
        if conn:
            cursor = conn.cursor()
            res = MISSING
            try:
                cursor.execute(sql_query)
                res = cursor.fetchmany(size)
            except sqlErr as err:
                log.error(f"failed fetchable query: {sql_query}\n{err}", exc_info=True)
                print(f"failed fetchable query: {sql_query}")
                traceback.print_exc()
            finally:
                cursor.close()
                conn.close()
            return res
        else:
            return



class Administration():
    """
    MySQL administration simplification
    -------
    \n
    works as `connect->execute preset queries->commit if necessary->disconnect`\n
    automates some boring processes like database or user creation for example\n
    --------------------------\n
    dbconfig: your database config as `dict`{\n
        `'host'`: host address as str (must be same as in mysql_config),\n
        `'user'`: db primary user username,\n
        `'passwd'`: db primary user password,\n
        `'user2'`: db secondary user username,\n
        `'passwd2'`: db secondary user password,\n
        `'database'`: your database name,\n
        `'tables'`: db tables structure presented as ['table (column TYPE, column TYPE, etc.)', etc.]\n
        }\n
    --------------------------\n
    mysql_config: your deafult mysql config as `dict`{\n
        `'host'`: host address as str,\n
        `'user'`: db root or admin user username,\n
        `'passwd'`: db root or admin user password,\n
        }\n
    this arg is required only when you set 'root' as True\n
    --------------------------\n
    alt_user: specify if you need to set this parameter to whole class instance
    """
    
    def __init__(self, dbconfig: dict, mysql_config: dict=MISSING, alt_user=False):
        self.dbconfig=dbconfig
        self.mysql_config=mysql_config
        self.alt_user=alt_user
    
    def create_db(self, add_users=False, add_tables=True):
        """
        creates a new database\n
        `your MySQL user must have required privileges and grant option to use this one!`\n
        ----------\n
        add_users: if you need to create new users for using this db (deafult is False)\n
        this will create 2 users, `their usernames and passwords must be specified in 'dbconfig' dict which is being given as a class arg`\n
        ------------\n
        add_tables: if you need to create some tables in this db (you probably do, right?..)\n
        tables and their structures `must be specified must be specified in 'dbconfig' dict which is being given as a class arg`
        """
        if add_users:
            conn = BasicAction().connect(self.dbconfig, root=True, rootconfig=self.mysql_config)
        else:
            conn = BasicAction().connect(self.dbconfig)
            if not conn:
                conn = BasicAction().connect(self.dbconfig, root=True, rootconfig=self.mysql_config)
        if conn:
            cursor=conn.cursor()
            try:
                cursor.execute(f"CREATE DATABASE {self.dbconfig['database']}")
                conn.commit()
            except sqlErr as err:
                log.error(f"failed creating database: {self.dbconfig['database']}\n{err}", exc_info=True)
                print(f"failed creating database: {self.dbconfig['database']}")
                traceback.print_exc()
            except Exception as err:
                log.error(f"unexpected error while creating database\n{err}", exc_info=True)
                print(f"unexpected error while creating database")
                traceback.print_exc()
            cursor.execute(f"USE {self.dbconfig['database']}")
            conn.commit()
            if add_tables:
                try:
                    self.create_tables(connection=conn)
                except:
                    pass
            if add_users:
                try:
                    self.create_users(connection=conn)
                except:
                    pass
    
    def create_users(self, connection=MISSING):
        """
        creates 2 new MySQL users which parameters (usernames and passwords) must be specified in 'dbconfig' dict which is being given as a class arg\n
        -----------\n
        connection: MySQL connection instance if you already have one (optional)\n
        `creates its own one as yor deafult MySQL user if not presented `
        """
        if connection is MISSING:
            conn = BasicAction().connect(self.dbconfig, root=True, rootconfig=self.mysql_config)
        else:
            conn = connection
        if conn:
            cursor = conn.cursor()
            try:
                cursor.execute(f"CREATE USER '{self.dbconfig['user']}'@'{self.dbconfig['host']}' IDENTIFIED BY '{self.dbconfig['passwd']}'")
                conn.commit()
                cursor.execute(f"GRANT ALL ON *.* TO '{self.dbconfig['user']}'@'{self.dbconfig['host']}' WITH GRANT OPTION")
                conn.commit()
                cursor.execute(f"FLUSH PRIVILEGES")
                conn.commit()
            except sqlErr as err:
                log.error(f"failed creating user: {self.dbconfig['user']}\n{err}", exc_info=True)
                print(f"failed creating user: {self.dbconfig['user']}")
                traceback.print_exc()
            except Exception as err:
                log.error(f"unexpected error while creating user\n{err}", exc_info=True)
                print(f"unexpected error while creating user")
                traceback.print_exc()
            try:
                cursor.execute(f"CREATE USER '{self.dbconfig['user2']}'@'{self.dbconfig['host']}' IDENTIFIED BY '{self.dbconfig['passwd2']}'")
                conn.commit()
                cursor.execute(f"GRANT ALL ON *.* TO '{self.dbconfig['user2']}'@'{self.dbconfig['host']}' WITH GRANT OPTION")
                conn.commit()
                cursor.execute(f"FLUSH PRIVILEGES")
                conn.commit()
            except sqlErr as err:
                log.error(f"failed creating user: {self.dbconfig['user2']}\n{err}", exc_info=True)
                print(f"failed creating user: {self.dbconfig['user2']}")
                traceback.print_exc()
            except Exception as err:
                log.error(f"unexpected error while creating user\n{err}", exc_info=True)
                print(f"unexpected error while creating user")
                traceback.print_exc()
            cursor.close()
            if connection is MISSING:
                conn.close()
    
    def create_tables(self, connection=MISSING, database: str=MISSING, table_listing: list=MISSING):
        """
        creates tables in your database\n
        -----------\n
        connection: MySQL connection instance if you already have one (optional)\n
        ---------------\n
        database: database to create tables in it (optional)\n
        `inherited from a class instance if not presented`\n
        --------------\n
        table_listing: db tables structure presented as ['table (column TYPE, column TYPE, etc.)', etc.] (optional)\n
        `inherited from a class instance if not presented`
        """
        if connection is MISSING:
            conn = BasicAction().connect(self.dbconfig, root=True, rootconfig=self.mysql_config)
        else:
            conn = connection
        if conn:
            cursor = conn.cursor()
            if database is MISSING:
                database = self.dbconfig['database']
            if table_listing is MISSING:
                table_listing = self.dbconfig['tables']
            if connection is MISSING:
                cursor.execute(f"USE {database}")
                conn.commit()
            else:
                db_old = conn._database
                cursor.execute(f"USE {database}")
                conn.commit()
            for table in table_listing:
                try:
                    cursor.execute(f"CREATE TABLE {table}")
                    conn.commit()
                except sqlErr as err:
                    log.error(f"{database} - failed creating table: {table}\n{err}", exc_info=True)
                    print(f"{database} - failed creating table: {table}")
                    traceback.print_exc()
                except Exception as err:
                    log.error(f"unexpected error while creating database\n{err}", exc_info=True)
                    print(f"unexpected error while creating database")
                    traceback.print_exc()
            if connection is MISSING:
                cursor.close()
                conn.close()
            else:
                cursor.execute(f"USE {db_old}")
                conn.commit()
                cursor.close()



class BasicAction():
    """
    necessary backend actions which you shouldn't call without extreme need
    -----------\n
    it may cause some global problems with your MySQL server or client
    """

    def __init__(self, alt_user=False):
        self.alt_user=alt_user

    def connect(self, dbconfig:dict, root=False, rootconfig: dict=MISSING, secondary: bool=MISSING):
        """
        `NEVER FUCKING CALL THIS INSIDE YOUR CODE`
        ------------
        \n
        otherwise you may cause MySQL connections excess!\n
        ---------\n
        connects you to database\n
        ---------\n
        root: set True if you need to use root user (no database by deafult)\n
        rootconfig: your deafult mysql config as `dict`{\n
        `'host'`: host address as str,\n
        `'user'`: db root or admin user username,\n
        `'passwd'`: db root or admin user password,\n
        }\n
        this arg is required only when you set 'root' as True\n
        --------------------------\n
        secondary: specify if you need to use a secondary user for this call only
        """
        if not root:
            if secondary is MISSING:
                secondary=self.alt_user
            try:
                connection_db = mysql.connector.connect(
                    host = dbconfig['host'],
                    user = dbconfig['user'] if not secondary else dbconfig['user2'],
                    passwd = dbconfig['passwd'] if not secondary else dbconfig['passwd2'],
                    database = dbconfig['database']
                )
                connection_db.commit()
                return connection_db
            except sqlErr as db_connection_error:
                log.error(db_connection_error, exc_info=True)
                #print("Возникла ошибка MySQL: ", db_connection_error)
                return
            except Exception as err:
                log.error(f"passing an unhandled exception:\n{err}", exc_info=True)
                print("passing an unhandled exception:")
                traceback.print_exc()
                return
        else:
            if rootconfig is MISSING:
                raise ValueError("You need to specify a 'rootconfig' arg while calling a root connection!")
            else:
                try:
                    connection_db = mysql.connector.connect(
                        host = rootconfig['host'],
                        user = rootconfig['user'],
                        passwd = rootconfig['passwd']
                    )
                    connection_db.commit()
                    return connection_db
                except sqlErr as db_connection_error:
                    log.error(db_connection_error, exc_info=True)
                    #print("Возникла ошибка MySQL: ", db_connection_error)
                    return
                except Exception as err:
                    log.error(f"passing an unhandled exception:\n{err}", exc_info=True)
                    print("passing an unhandled exception:")
                    traceback.print_exc()
                    return



class _Clear:
    """MySQL response clearer"""

    def string(obj):
        """clears response from `fetchone`"""
        chars_to_remove = ["(", ")", ",", "'"]
        element = str()
        try:
            for element in obj:
                for char in chars_to_remove:
                    element = str(element)
                    element = element.replace(char, "")
        except:
            pass
        return element
    
    def listing(obj):
        """clears response from `fetchall`"""
        chars_to_remove = ["(", ")", ",", "'"]
        clear_list = []
        try:
            for element in obj:
                for char in chars_to_remove:
                    element = str(element)
                    element = element.replace(char, "")
                clear_list.append(element)
        except:
            pass
        return clear_list


from typing import Any

import pathlib



__all__ = [
    'MISSING',
    'ROOT',
    'MAXINT'
]



ROOT = pathlib.Path(pathlib.Path.cwd())
MAXINT = 2147483647



class _MissingSentinel():
    def __eq__(self, other: Any) -> bool:
        return False

    def __hash__(self) -> int:
        return 0

    def __bool__(self) -> bool:
        return False

    def __repr__(self) -> str:
        return "..."

MISSING: Any = _MissingSentinel()



def _yesornot(text):
    from sys import setrecursionlimit as recurlim
    recurlim(MAXINT)
    """
    repeats question until user answers `yes` or `no`
    """
    print(text, ' (y/n) ', end='')
    answer = input()
    if answer=='y' or answer=='Y' or answer=='yes' or answer=='Yes' or answer=='YES':
        return True
    elif answer=='n' or answer=='N' or answer=='no' or answer=='No' or answer=='NO':
        return False
    else:
        return _yesornot(text)